import psutil
import math
import requests

bytes_per_GB = 1024**3

def convert_bytes_to_GB(x_bytes):
    x_GB = x_bytes / bytes_per_GB
    return x_GB

def convert_GB_to_bytes(x_GB):
    x_bytes = x_GB * bytes_per_GB
    return x_bytes

def get_RAM_available_in_bytes():
    return psutil.virtual_memory()[1]

def get_RAM_available_in_GB():
    return convert_bytes_to_GB(get_RAM_available_in_bytes())

def get_RAM_used_in_bytes():
    return psutil.virtual_memory()[3]

def get_RAM_used_in_GB():
    return convert_bytes_to_GB(get_RAM_used_in_bytes())

def round_down_n_decimal_places(a, n):
    return math.floor(a * 10**n) / 10**n

def download_file_from_google_drive(id, destination):
    # link: https://stackoverflow.com/a/39225272/5040208
    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params = { 'id' : id , 'confirm': 1 }, stream = True)
    token = get_confirm_token(response)

    if token:
        params = { 'id' : id, 'confirm' : token }
        response = session.get(URL, params = params, stream = True)

    save_response_content(response, destination)

def get_confirm_token(response):
    # link: https://stackoverflow.com/a/39225272/5040208
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None

def save_response_content(response, destination):
    # link: https://stackoverflow.com/a/39225272/5040208
    CHUNK_SIZE = 32768

    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk: # filter out keep-alive new chunks
                f.write(chunk)



import time
import queue
import sys
from collections import namedtuple

import numpy as np

from .node import Node, branch  # , presolve


from .solvers import (
    sparseLogRegModel_big_n as sparseLogRegModel_big_n_with_cache,
    total_size,
)

import psutil
import gc
import scipy

from .utils import (
    get_RAM_available_in_bytes,
    get_RAM_available_in_GB,
    get_RAM_used_in_bytes,
    get_RAM_used_in_GB,
)


class DataClass:
    def __init__(self, X, y, lambda2):
        """Create a DataClass object used to store data for the BnBTree class

        Args:
            X (np.array): n x p numpy array of features
            y (np.array): 1 dimensional numpy array of size n of predictions
            lambda2 (float): coefficient of l2 regularization
        """
        self.p = X.shape[1]
        self.n = X.shape[0]
        self.X = X
        self.y = y
        self.X_norm_2 = np.linalg.norm(X, axis=0) ** 2
        self.XTX = X.T @ X
        # self.X_norm_2 = self.XTX.diagonal() ** 2
        self.XTX_lambda2 = self.XTX + lambda2 * np.eye(N=self.p)
        self.XTy = y.dot(self.X)
        self.lambda2 = lambda2
        self.half_Lipschitz = self.XTX_lambda2.diagonal()

        eigenvalues = np.linalg.eigvalsh(self.XTX)
        self.smallest_eigenval = np.linalg.eigvalsh(self.XTX)[0] * 0.999
        # print("we take lambda to be", self.smallest_eigenval)

        self.rho_ADMM = 2 / np.sqrt(
            (eigenvalues[1] - self.smallest_eigenval)
            * (eigenvalues[-1] - self.smallest_eigenval)
        )

        self.rho_ADMM_is_finetuned = True
        if eigenvalues[1] < self.smallest_eigenval + 1e-6:
            self.rho_ADMM_is_finetuned = False

        if self.smallest_eigenval < 1e-3:
            self.smallest_eigenval = 0.0
        self.XTX_minus_smallest_eigenval = (
            self.XTX - np.eye(self.p) * self.smallest_eigenval
        )
        self.lambda2_plus_smallest_eigenval = self.lambda2 + self.smallest_eigenval


class BNBTree:
    def __init__(
        self,
        X,
        y,
        int_tol=1e-4,
        gap_tol=1e-4,
        lambda2=1e-5,
        max_memory_GB=300,
        useBruteForce=False,
        tighten_bound_via_ADMM=True,
    ):
        """Initialize the BnBTree class with the data and parameters

        Args:
            X (np.array): n x p numpy array of features
            y (np.array): 1 dimensional numpy array of size n of predictions
            int_tol (float, optional): integer tolerance hyperparameter. Defaults to 1e-4.
            gap_tol (float, optional): optimality gap tolerance hyperparameter. Defaults to 1e-4.
            lambda2 (float, optional): coefficient of l2 regularization. Defaults to 1e-5.
            max_memory_GB (int, optional): max memory to use to store all unprocessed nodes and heuristic solutions. Defaults to 300.
            useBruteForce (bool, optional): use brute force search when the number of enumeration is small. Defaults to False.
            tighten_bound_via_ADMM (bool, optional): whether to use ADMM to tighten the lower bound computed by the Fast Solve method. Defaults to True.
        """
        data = DataClass(X, y, lambda2)
        self.data = data

        self.int_tol = int_tol
        self.gap_tol = gap_tol

        self.useBruteForce = useBruteForce
        self.bruteForceThreshold = 1000

        self.tighten_bound_via_ADMM = tighten_bound_via_ADMM

        self.bfs_queue = None
        self.dfs_queue = None

        self.levels = {}
        self.number_of_nodes = 0

        self.root = None

        available_memory_GB = get_RAM_available_in_GB()
        if max_memory_GB is None:
            print(
                "No max_memory_GB is given. Using all available memory ({} GB) in the machine".format(
                    available_memory_GB
                )
            )
            self.max_memory_GB = available_memory_GB
        elif max_memory_GB > available_memory_GB:
            print(
                "max_memory_GB is larger than available memory. Using all available memory ({} GB) in the machine".format(
                    available_memory_GB
                )
            )
            self.max_memory_GB = available_memory_GB
        else:
            print("Using max memory ({} GB)".format(max_memory_GB))
            self.max_memory_GB = max_memory_GB
        self.safe_max_memory_GB = 0.95 * self.max_memory_GB
        self.RAM_used_GB_start = get_RAM_used_in_GB()

        self.upper_solver_with_cache = sparseLogRegModel_big_n_with_cache(
            data=self.data,
            parent_size=50,
            child_size=50,
            max_memory_GB=50,
        )

    def get_RAM_used_since_start(self):
        RAM_used_GB = get_RAM_used_in_GB()
        return RAM_used_GB - self.RAM_used_GB_start

    def solve(
        self,
        k,
        gap_tol=1e-2,
        number_of_dfs_levels=0,
        verbose=False,
        time_limit=3600,
    ):
        """Solve the k-sparse ridge regression problem using branch and bound

        Args:
            k (int): cardinality constraint
            gap_tol (float, optional): optimality gap tolerance hyperparameter. Defaults to 1e-2.
            number_of_dfs_levels (int, optional): number of levels for depth-first search during branch and bound. Defaults to 0.
            verbose (bool, optional): whether to print informations into terminal. Defaults to False.
            time_limit (int, optional): time limit (in seconds) of running branch and bound. Defaults to 3600.

        Returns:
            float: cost or loss of the best solution found
            np.array: best solution found
            float: best gap found
            float: best lower bound found
            float: running time of branch and bound
        """

        st = time.time()
        upper_beta = np.zeros((self.data.p,))
        upper_bound = self.data.XTX_lambda2.dot(upper_beta).dot(upper_beta)

        # if verbose:
            # print(f"initializing took {time.time() - st} seconds")

        st = time.time()
        # root node
        self.root = Node(None, None, None, data=self.data)

        self.bfs_queue = queue.Queue()
        self.dfs_queue = queue.LifoQueue()
        self.bfs_queue.put(self.root)

        # lower and upper bounds initialization
        dual_bound = {}
        self.levels = {0: 1}
        min_open_level = 0

        max_lower_bound_value = -sys.maxsize
        best_gap = gap_tol + 1

        if verbose:
            if number_of_dfs_levels > 0:
                print("using depth-first search for the first {} levels".format(number_of_dfs_levels))
            else:
                print("using breadth-first search")

        upper_bound = self.root.upper_solve_with_cache(k, self.upper_solver_with_cache)
        upper_beta = self.root.upper_beta.copy()
        if (k == 1) or (time_limit < 0):
            max_lower_bound_value = upper_bound
            best_gap = 0
            return (
                upper_bound,
                upper_beta,
                best_gap,
                max_lower_bound_value,
                time.time() - st,
            )
        
        # print("initial upper_bound is", upper_bound)
        # print("initial nonzero indices are", np.nonzero(upper_beta)[0])

        if self.root.data.rho_ADMM_is_finetuned is False:
            self.root.finetune_ADMM_rho(k, upper_bound, factor=10)

        RAM_used_GB_since_start = self.get_RAM_used_since_start()

        print(
            f"'l' -> level(depth) of BnB tree, ",
            f"'d' -> best dual bound, ",
            f"'u' -> best upper(primal) bound, ",
            f"'g' -> optimiality gap, ",
            f"'t' -> time"
        )

        # keep searching through the queue if the queue is not empty AND time limit is not reached
        while (
            (self.bfs_queue.qsize() > 0 or self.dfs_queue.qsize() > 0)
            and (time.time() - st < time_limit)
            and (RAM_used_GB_since_start < self.safe_max_memory_GB)
        ):

            gc.collect()

            RAM_used_GB_since_start = self.get_RAM_used_since_start()

            # get current node
            if self.dfs_queue.qsize() > 0:
                curr_node = self.dfs_queue.get()
            else:
                curr_node = self.bfs_queue.get()

            # prune?
            if (
                curr_node.parent_lower_bound
                and upper_bound <= curr_node.parent_lower_bound
            ):
                self.levels[curr_node.level] -= 1
                continue

            # calculate dual values
            curr_dual = self._solve_node(
                curr_node, k, dual_bound, upper_bound, tighten_bound_via_ADMM=self.tighten_bound_via_ADMM
            )

            # prune?with newly calculated lower bound
            if curr_dual >= upper_bound:
                curr_node.delete_storedData_on_allowed_support()
                continue

            curr_upper_bound = curr_node.upper_solve_with_cache(
                k, self.upper_solver_with_cache
            )

            # if verbose:
            #     print("curr_upper_bound: {}, curr_dual: {}".format(curr_upper_bound, curr_dual)) # debugging

            if curr_upper_bound < upper_bound:
                upper_bound = curr_upper_bound
                upper_beta = curr_node.upper_beta.copy()
                best_gap = (upper_bound - max_lower_bound_value) / abs(upper_bound)
                # print("********************************\nfind better solution with upper_bound", upper_bound)
                # print("the nonzero indices are", np.nonzero(upper_beta)[0], "\n********************************")

            # update gap?
            if self.levels[min_open_level] == 0:
                # print(
                #     "there are {} nodes left in the bfs queue".format(
                #         self.bfs_queue.qsize()
                #     )
                # )
                del self.levels[min_open_level]
                max_lower_bound_value = max(
                    [j for i, j in dual_bound.items() if i <= min_open_level]
                )
                best_gap = (upper_bound - max_lower_bound_value) / abs(upper_bound)
                if verbose:
                    print(
                        # f"l: {min_open_level}, d: {max_lower_bound_value}, "
                        # f"u: {upper_bound}, g: {best_gap}, "
                        # f"t: {time.time() - st} s"
                        "l: {}, ".format(min_open_level).ljust(8),
                        "d: {:.10f}, ".format(max_lower_bound_value).ljust(25),
                        "u: {:.10f}, ".format(upper_bound).ljust(25),
                        "g: {:.10f}, ".format(best_gap).ljust(15),
                        "t: {:.5f} s".format(time.time() - st).ljust(12)
                    )
                min_open_level += 1

            # arrived at a solution?
            if best_gap <= gap_tol:
                return (
                    upper_bound,
                    upper_beta,
                    best_gap,
                    max_lower_bound_value,
                    time.time() - st,
                )

            # branch?
            curr_gap = (curr_upper_bound - curr_dual) / abs(curr_upper_bound)
            if curr_gap <= gap_tol:
                # print("curr_gap is smaller than gap_tol; skipping!")
                pass
            elif (curr_dual < upper_bound) and (
                len(curr_node.fixed_support_on_allowed_support) < k
            ):
                left_node, right_node = branch(curr_node, k)
                self.levels[curr_node.level + 1] = (
                    self.levels.get(curr_node.level + 1, 0) + 2
                )
                if curr_node.level < min_open_level + number_of_dfs_levels:
                    self.dfs_queue.put(right_node)
                    self.dfs_queue.put(left_node)
                else:
                    self.bfs_queue.put(right_node)
                    self.bfs_queue.put(left_node)
            else:
                # print(
                #     "fixed support size is {}".format(
                #         len(curr_node.fixed_support_on_allowed_support)
                #     )
                # )
                # print("no branching!!!")
                pass

            curr_node.delete_storedData_on_allowed_support()

        if not (self.bfs_queue.qsize() > 0 or self.dfs_queue.qsize() > 0):
            # print("There are no nodes left in the queue")
            # update lower bound and gap
            max_lower_bound_value = upper_bound
            best_gap = 0.0
        elif RAM_used_GB_since_start >= self.safe_max_memory_GB:
            print(
                "RAM used since start is greater than 0.95*max_memory(0.95*{} GB = {} GB)!".format(
                    self.max_memory_GB, self.safe_max_memory_GB
                )
            )
        else:
            print("Time limit is reached!")
        return (
            upper_bound,
            upper_beta,
            best_gap,
            max_lower_bound_value,
            time.time() - st,
        )

    def _solve_node(self, curr_node, k, dual_, upper_bound, tighten_bound_via_ADMM=True):
        """Solve the perspective relaxation of the node and update the dual bound

        Args:
            curr_node (CustomClass): current node to be solved
            k (int): cardinality constraint
            dual_ (list): list of minimum dual values at each depth of the tree
            upper_bound (float): current loss of the best solution found for this node
            tighten_bound_via_ADMM (bool, optional): whether to refine the dual bound computed by the Fast Solve method. Defaults to True.

        Returns:
            float: dual value of the perspective relaxation of the current node
        """
        self.number_of_nodes += 1
        total_num_bruteForce = scipy.special.comb(
            len(curr_node.unfixed_support_on_allowed_support),
            k - len(curr_node.fixed_support_on_allowed_support),
        )
        if self.useBruteForce and (total_num_bruteForce < self.bruteForceThreshold):
            curr_dual = curr_node.lower_solve_brute_force(
                k, self.upper_solver_with_cache
            )
            # print("lower bound by brute force method is", curr_dual)
        else:
            curr_dual = curr_node.lower_solve_fast(k, upper_bound)
            # print("lower bound by fast method is", curr_dual)

            if tighten_bound_via_ADMM and (curr_dual < upper_bound):
                # curr_dual = curr_node.lower_solve_improve(k, upper_bound)
                curr_dual = curr_node.lower_solve_admm(k, upper_bound)
                # print("lower bound by admm method is", curr_dual)

        dual_[curr_node.level] = min(curr_dual, dual_.get(curr_node.level, sys.maxsize))
        self.levels[curr_node.level] -= 1
        return curr_dual

    def solve_root(
        self,
        k,
        gap_tol=1e-2,
        verbose=False,
    ):
        # should I keep this function?

        st = time.time()
        upper_beta = np.zeros((self.data.p,))
        upper_bound = self.data.XTX_lambda2.dot(upper_beta).dot(upper_beta)

        st = time.time()
        # root node
        self.root = Node(None, None, None, data=self.data)

        dual_bound = {}
        self.levels = {0: 1}
        max_lower_bound_value = -sys.maxsize
        best_gap = gap_tol + 1

        # calculate upper bound
        upper_bound = self.root.upper_solve_with_cache(k, self.upper_solver_with_cache)
        upper_beta = self.root.upper_beta.copy()

        # calculate lower bound
        curr_dual = self._solve_node(
            self.root, k, dual_bound, upper_bound, tighten_bound_via_ADMM=self.tighten_bound_via_ADMM
        )
        max_lower_bound_value = curr_dual
        best_gap = (upper_bound - max_lower_bound_value) / abs(upper_bound)

        if k == 1:
            max_lower_bound_value = upper_bound
            best_gap = 0

        return (
            upper_bound,
            upper_beta,
            best_gap,
            max_lower_bound_value,
            time.time() - st,
        )


import numpy as np
import sys
import functools

import gc
import psutil

from .utils import convert_GB_to_bytes

from sys import getsizeof, stderr
from itertools import chain
from collections import deque
try:
    from reprlib import repr
except ImportError:
    pass

from itertools import combinations
import scipy

def total_size(o, handlers={}, verbose=False):
    """ Returns the approximate memory footprint an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}

    """
    dict_handler = lambda d: chain.from_iterable(d.items())
    all_handlers = {tuple: iter,
                    list: iter,
                    deque: iter,
                    dict: dict_handler,
                    set: iter,
                    frozenset: iter,
                   }
    all_handlers.update(handlers)     # user handlers take precedence
    seen = set()                      # track which object id's have already been seen
    default_size = getsizeof(0)       # estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:       # do not double count the same object
            return 0
        seen.add(id(o))
        s = getsizeof(o, default_size)

        if verbose:
            print(s, type(o), repr(o), file=stderr)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)

def calculate_size(tmp_dict):
    size = 0
    for tmp_key in tmp_dict:
        size += (sys.getsizeof(tmp_dict[tmp_key][0]) + sys.getsizeof(tmp_dict[tmp_key][1]))
    return size



class linRegModel_unnormalized_big_n:
    def __init__(self, data):
        """Initialize the linear regression model

        Args:
            data (CustomClass): data containing all the necessary information
        """
        self.X = data.X
        self.n, self.p = data.n, data.p

        self.y = data.y.reshape(-1).astype(float)
        self.XTX_lambda2 = data.XTX_lambda2
        self.XTy = data.XTy
        self.betas = np.zeros((self.p, ))

        self.r = self.XTX_lambda2.dot(self.betas)
        self.loss = self.r.dot(self.betas) - 2 * self.XTy.dot(self.betas)

        self.lambda2 = data.lambda2

        self.half_Lipschitz = data.half_Lipschitz
        self.total_child_added = 0
    
    def warm_start_from_betas(self, betas):
        """Warm start the model from a given betas

        Args:
            betas (np.array): 1D array of coefficients
        """
        self.betas = betas
        self.r = self.XTX_lambda2.dot(self.betas)
        self.total_child_added = 0

    def get_betas(self):
        """Get the current betas

        Returns:
            np.array: 1D array of coefficients
        """
        return self.betas

    def get_betas_r(self):
        """Get the current betas, and r

        Returns:
            np.array: 1D array of coefficients
            np.array: 1D array of intermediate values - XTX_lambda2.dot(betas)
        """
        return self.betas, self.r
    
    def get_betas_r_loss(self):
        """Get the current betas, r, and loss

        Returns:
            np.array: 1D array of coefficients
            np.array: 1D array of intermediate values - XTX_lambda2.dot(betas)
            float: loss
        """
        return self.betas, self.r, self.loss
        
    # @functools.cache
    def finetune_on_current_support(self, supp_mask):
        """Finetune the current solution on a given support

        Args:
            supp_mask (np.array): 1D array of boolean values indicating the support

        Returns:
            np.array: 1D array of coefficients on the support
            np.array: 1D array of intermediate values - XTX_lambda2.dot(betas)
            float: loss
        """

        XTX_lambda2_supp_mask = self.XTX_lambda2[supp_mask]
        XTy_supp_mask = self.XTy[supp_mask]
        betas_on_supp = np.linalg.solve(XTX_lambda2_supp_mask[:, supp_mask], XTy_supp_mask)
        r = betas_on_supp.dot(XTX_lambda2_supp_mask)
        loss = r[supp_mask].dot(betas_on_supp) - 2 * XTy_supp_mask.dot(betas_on_supp)

        return betas_on_supp, r, loss


class sparseLogRegModel_big_n(linRegModel_unnormalized_big_n):
    def __init__(self, data, parent_size=10, child_size=10, allowed_supp_mask=None, max_memory_GB=50):
        """Initialize the sparse logistic regression model

        Args:
            data (CustomClass): data containing all the necessary information
            parent_size (int, optional): number of solutions to keep after doing beam search so that these can be used as parent solutions for the next stage of support expansion. Defaults to 10.
            child_size (int, optional): number of solutions to explore for each parent solution during beam search. Defaults to 10.
            allowed_supp_mask (np.array, optional): 1D of boolean values indicating which features are allowed to use. Defaults to None.
            max_memory_GB (int, optional): maximum memory (in GB) to use. Defaults to 50.
        """
        super().__init__(data=data)
        
        self.parent_size = parent_size
        self.child_size = child_size

        self.supp_mask_arr_parent = np.zeros((parent_size, self.p), dtype=bool)
        self.num_parent = 1

        self.total_child_size = self.parent_size * self.child_size
        self.supp_mask_arr_child = np.zeros((self.total_child_size, self.p), dtype=bool)
        self.loss_arr_child = 1e12 * np.ones((self.total_child_size, ))

        if allowed_supp_mask is None:
            self.allowed_supp_mask = np.ones((self.p, ), dtype=bool)
        else:
            self.allowed_supp_mask = allowed_supp_mask
        
        self.saved_solution = {}
        supp_mask_all_False = np.zeros((self.p, ), dtype=bool)
        tmp_support_str = supp_mask_all_False.tobytes()
        self.saved_solution[tmp_support_str] = (None, self.r, 0.0)

        self.max_memory_GB = max_memory_GB
        tmp_vec_bytes = sys.getsizeof(np.zeros((self.p, )))
        self.max_saved_solutions = int(convert_GB_to_bytes(self.max_memory_GB) / tmp_vec_bytes)

        # print("max number of saved solutions is", self.max_saved_solutions)

    def reset_fixed_supp_and_allowed_supp(self, fixed_supp_mask, allowed_supp_mask):
        """Reset the fixed support and allowed support

        Args:
            fixed_supp_mask (np.array): 1D array of boolean values indicating the fixed support
            allowed_supp_mask (np.array): 1D array of boolean values indicating the allowed support
        """
        self.fixed_supp_mask = fixed_supp_mask
        self.allowed_supp_mask = allowed_supp_mask
        self.betas.fill(0)
        tmp_support_str = fixed_supp_mask.tobytes()
        if tmp_support_str in self.saved_solution:
            betas_on_supp_tmp, self.r, self.loss = self.saved_solution[tmp_support_str]
        else:
            betas_on_supp_tmp, self.r, self.loss = self.finetune_on_current_support(fixed_supp_mask)
        self.betas[fixed_supp_mask] = betas_on_supp_tmp
    
    def get_sparse_sol_via_brute_force(self, k):
        """Get sparse solution through brute force

        Args:
            k (int): cardinality of the final sparse solution
        """

        # enumerate all possible support sets
        best_loss = self.loss
        best_supp_str = self.fixed_supp_mask.tobytes()

        total_num_combinations = scipy.special.comb(sum(self.allowed_supp_mask) - sum(self.fixed_supp_mask), k - sum(self.fixed_supp_mask), exact=True)
        total_supp_mask = np.zeros((total_num_combinations, self.p), dtype=bool)
        total_supp_mask[:, self.fixed_supp_mask] = True

        additional_fixed_supp_masks = combinations(list(np.where(np.logical_xor(self.allowed_supp_mask, self.fixed_supp_mask))[0]), k - sum(self.fixed_supp_mask))

        for i, additional_fixed_supp_mask in enumerate(additional_fixed_supp_masks):
            total_supp_mask[i, list(additional_fixed_supp_mask)] = True
            tmp_support_str = total_supp_mask[i].tobytes()
            if tmp_support_str in self.saved_solution:
                betas_on_supp_tmp, r_on_supp_tmp, loss_tmp = self.saved_solution[tmp_support_str]
            else:
                betas_on_supp_tmp, r_on_supp_tmp, loss_tmp = self.finetune_on_current_support(total_supp_mask[i])
                if len(self.saved_solution) < self.max_saved_solutions:
                    self.saved_solution[tmp_support_str] = (betas_on_supp_tmp, r_on_supp_tmp, loss_tmp)

            if loss_tmp < best_loss:
                best_loss = loss_tmp
                best_supp_str = tmp_support_str
        
        self.betas.fill(0)
        best_betas_on_supp, self.r, self.loss = self.finetune_on_current_support(np.frombuffer(best_supp_str, dtype=bool))
        self.betas[np.frombuffer(best_supp_str, dtype=bool)] = best_betas_on_supp

    def get_sparse_sol_via_OMP(self, k):
        """Get sparse solution through beam search and orthogonal matching pursuit (OMP), for level i, each parent solution generates [child_size] child solutions, so there will be [parent_size] * [child_size] number of total child solutions. However, only the top [parent_size] child solutions are retained as parent solutions for the next level i+1

        Args:
            k (int): cardinality of the final sparse solution
        """

        nonzero_indices_set = set(np.where(np.abs(self.betas) > 1e-9)[0])
        num_nonzero = len(nonzero_indices_set)
        zero_indices_set = set(np.where(self.allowed_supp_mask)[0]) - nonzero_indices_set

        if len(zero_indices_set) == 0:
            return
    
        self.supp_mask_arr_parent[0] = np.abs(self.betas) > 1e-9

        self.num_parent = 1
        self.forbidden_support = set()

        while num_nonzero < min(k, self.p):
            num_nonzero += 1
            self.beamSearch_multipleSupports_via_OMP_by_1()
        
        del self.forbidden_support
        gc.collect()

        best_sol_supp_mask = self.supp_mask_arr_parent[0]
        betas_on_supp, self.r, self.loss = self.finetune_on_current_support(best_sol_supp_mask)
        self.betas.fill(0.0)
        self.betas[best_sol_supp_mask] = betas_on_supp
   
    def beamSearch_multipleSupports_via_OMP_by_1(self):
        """Each parent solution generates [child_size] child solutions, so there will be [parent_size] * [child_size] number of total child solutions. However, only the top [parent_size] child solutions are retained as parent solutions for the next level i+1.
        """
        self.loss_arr_child.fill(1e32)
        self.total_child_added = 0

        for i in range(self.num_parent):
            self.expand_parent_i_support_via_OMP_by_1(i)
        
        child_indices = np.argsort(self.loss_arr_child)[:min(self.parent_size, self.total_child_added)] # get indices of children which have the smallest losses
        num_child_indices = len(child_indices)

        # print("size of saved solution dict is {}".format(total_size(self.saved_solution)))


        self.supp_mask_arr_parent[:num_child_indices] = self.supp_mask_arr_child[child_indices]

        self.num_parent = num_child_indices

    def expand_parent_i_support_via_OMP_by_1(self, i):
        """For parent solution i, generate [child_size] child solutions

        Args:
            i (int): index of the parent solution
        """
        fixed_supp_mask = self.supp_mask_arr_parent[i]
        unfixed_and_allowed_mask = np.logical_xor(self.allowed_supp_mask, fixed_supp_mask)
        unfixed_and_allowed_indicies = np.where(unfixed_and_allowed_mask)[0]


        tmp_support_str = self.supp_mask_arr_parent[i].tobytes()
        if tmp_support_str in self.saved_solution:
            _, r_parent_i, _ = self.saved_solution[tmp_support_str]
        else:
            betas_on_supp_tmp, r_parent_i, loss_tmp = self.finetune_on_current_support(self.supp_mask_arr_parent[i])
            if len(self.saved_solution) < self.max_saved_solutions:
                self.saved_solution[tmp_support_str] = (betas_on_supp_tmp, r_parent_i, loss_tmp)
        half_grad_on_unfixed_and_allowed_supp = r_parent_i[unfixed_and_allowed_indicies] - self.XTy[unfixed_and_allowed_indicies]
        abs_half_grad_on_unfixed_and_allowed_supp =  half_grad_on_unfixed_and_allowed_supp ** 2 / self.half_Lipschitz[unfixed_and_allowed_indicies]

        num_new_js = min(self.child_size, len(unfixed_and_allowed_indicies))
        new_js = unfixed_and_allowed_indicies[np.argsort(-abs_half_grad_on_unfixed_and_allowed_supp)][:num_new_js]
        child_start, child_end = i * self.child_size, i*self.child_size + num_new_js

        self.supp_mask_arr_child[child_start:child_end] = self.supp_mask_arr_parent[i]

        for l in range(num_new_js):
            child_id = child_start + l
            self.supp_mask_arr_child[child_id, new_js[l]] = True
            tmp_support_str = self.supp_mask_arr_child[child_id].tobytes()
            if tmp_support_str not in self.forbidden_support:
                self.total_child_added += 1
                self.forbidden_support.add(tmp_support_str)

                if tmp_support_str in self.saved_solution:
                    _, _, self.loss_arr_child[child_id] = self.saved_solution[tmp_support_str]
                else:
                    betas_on_supp_tmp, r_tmp, self.loss_arr_child[child_id] = self.finetune_on_current_support(self.supp_mask_arr_child[child_id])
                    if len(self.saved_solution) <= self.max_saved_solutions:
                        self.saved_solution[tmp_support_str] = (betas_on_supp_tmp, r_tmp, self.loss_arr_child[child_id])
            
            # print("forbidden_support has {} elements".format(len(self.forbidden_support)))
            # print("saved_solution has {} elements".format(len(self.saved_solution)))



from copy import deepcopy
import numpy as np
import math

import time
import sys

from sklearn.isotonic import IsotonicRegression
import scipy
import gc
import psutil

gc.enable()


def get_RAM_used_in_GB():
    return psutil.virtual_memory()[3] / 1000000000


class Node:
    def __init__(self, parent, zlb, zub, **kwargs):
        """Initialize a node in the branch and bound tree.

        Args:
            parent (CustomClass): parent node
            zlb (np.array): 1D array indicating lower bound of each binary variable
            zub (np.array): 1D array indicating upper bound of each binary variable
        """

        self.data = kwargs.get("data", parent.data if parent else None)
        self.inherit_parent_upper_sol = kwargs.get("inherit_parent_upper_sol", False)
        self.inherit_parent_lower_sol = kwargs.get("inherit_parent_lower_sol", False)
        if parent is not None:
            self.parent_lower_bound = parent.lower_bound
        else:
            self.parent_lower_bound = 0

        self.level = parent.level + 1 if parent else 0

        if zlb is not None:
            self.zlb = zlb
        else:
            self.zlb = np.zeros((self.data.p,), dtype=bool)
        if zub is not None:
            self.zub = zub
        else:
            self.zub = np.ones((self.data.p,), dtype=bool)

        if self.inherit_parent_upper_sol:
            self.upper_bound = parent.upper_bound
            self.upper_beta = parent.upper_beta
            self.upper_r = parent.upper_r
        else:
            self.upper_bound = None
            self.upper_beta = None
            self.upper_r = None

        if self.inherit_parent_lower_sol:
            # self.lower_bound = parent.lower_bound
            self.lower_beta_on_allowed_support = (
                parent.lower_beta_on_allowed_support.copy()
            )
        else:
            self.lower_bound = None
            self.lower_beta_on_allowed_support = None

        self.allowed_support = np.where(self.zub == True)[0]
        self.fixed_support_on_allowed_support = np.where(
            self.zlb[self.allowed_support] == True
        )[0]
        self.unfixed_support_on_allowed_support = np.where(
            self.zlb[self.allowed_support] == False
        )[0]

        self.data_isStored_on_allowed_support = False
        self.lower_solve_used_brute_force = False

    def delete_storedData_on_allowed_support(self):
        """Delete the stored data on the allowed support to save memory
        """
        RAM_used = get_RAM_used_in_GB()
        # print("before deleting, RAM_used is", RAM_used)
        del self.allowed_support_2d_indices
        del self.XTX_minus_smallest_eigenval_on_allowed_support
        del self.XTX_lambda2_on_allowed_support
        del self.XTy_on_allowed_support
        del self.allowed_support
        del self.fixed_support_on_allowed_support
        del self.unfixed_support_on_allowed_support
        gc.collect()
        # print("after deleting, RAM_used is", RAM_used)

    def store_data_on_allowed_support(self):
        """Store the data on the allowed support to save future computation
        """
        # self.XTX_minus_smallest_eigenval_on_allowed_support = self.data.XTX_minus_smallest_eigenval[self.allowed_support][:, self.allowed_support]
        self.allowed_support_2d_indices = np.ix_(
            self.allowed_support, self.allowed_support
        )
        self.XTX_minus_smallest_eigenval_on_allowed_support = (
            self.data.XTX_minus_smallest_eigenval[self.allowed_support_2d_indices]
        )

        self.XTX_lambda2_on_allowed_support = self.data.XTX_lambda2[
            self.allowed_support_2d_indices
        ]
        self.XTX_minus_smallest_eigenval_on_allowed_support_twice = (
            2 * self.XTX_lambda2_on_allowed_support
        )
        self.XTy_on_allowed_support = self.data.XTy[self.allowed_support]

        self.data_isStored_on_allowed_support = True

    def upper_solve(self, k, upper_solver, parent_size=50, child_size=50):
        """Solve the k-sparse ridge regression for the current node using heuristic method

        Args:
            k (int): cardinality constraint
            upper_solver (CustomClass): solver that can search for the k-sparse solution
            parent_size (int, optional): parent size of the beam search procedure. Defaults to 50.
            child_size (int, optional): child size of the beam search procedure. Defaults to 50.

        Returns:
            float: loss of the k-sparse solution
        """
        if self.data_isStored_on_allowed_support is False:
            self.store_data_on_allowed_support()

        # print("inherit parent upper sol is {}".format(self.inherit_parent_upper_sol))
        if self.inherit_parent_upper_sol:
            return self.upper_bound
        betas_warmstart = np.zeros((len(self.allowed_support),))
        if len(self.fixed_support_on_allowed_support) >= 1:
            sub_betas_warmstart = np.linalg.solve(
                self.data.XTX_lambda2[self.fixed_support_on_allowed_support][
                    :, self.fixed_support_on_allowed_support
                ],
                self.data.XTy[self.fixed_support_on_allowed_support],
            )
            betas_warmstart[self.fixed_support_on_allowed_support] = sub_betas_warmstart

        upper_solver.reset(
            self.data.X[:, self.allowed_support],
            self.data.X_norm_2[self.allowed_support],
            self.data.XTX_lambda2[self.allowed_support][:, self.allowed_support],
            self.data.XTy[self.allowed_support],
        )
        upper_solver.warm_start_from_betas(betas_warmstart)

        upper_solver.get_sparse_sol_via_OMP(
            k=k, parent_size=parent_size, child_size=child_size
        )

        (
            betas_on_allowed_support,
            r_on_allowed_support,
        ) = upper_solver.get_betas_r()

        self.upper_beta = np.zeros((self.data.p,))
        self.upper_beta[self.allowed_support] = betas_on_allowed_support

        self.upper_r = np.zeros((self.data.p,))
        self.upper_r[self.allowed_support] = r_on_allowed_support

        self.upper_bound = min(upper_solver.loss_arr_child)

        return self.upper_bound

    def upper_solve_with_cache(self, k, upper_solver):
        """Solve the k-sparse ridge regression for the current node using heuristic method

        Args:
            k (int): cardinality constraint
            upper_solver (CustomClass): solver that can search for the k-sparse solution

        Returns:
            float: loss of the k-sparse solution
        """
        if self.lower_solve_used_brute_force:
            return self.upper_bound

        if self.data_isStored_on_allowed_support is False:
            self.store_data_on_allowed_support()

        # print("inherit parent upper sol is {}".format(self.inherit_parent_upper_sol))
        if self.inherit_parent_upper_sol:
            return self.upper_bound

        # allowed_supp_mask = self.allowed_support
        # fixed_supp_mask = np.zeros((self.data.p, ), dtype=bool)
        allowed_supp_mask = self.zub == 1
        fixed_supp_mask = self.zlb == 1

        upper_solver.reset_fixed_supp_and_allowed_supp(
            fixed_supp_mask, allowed_supp_mask
        )

        upper_solver.get_sparse_sol_via_OMP(k=k)

        # betas_on_allowed_support, r_on_allowed_support = upper_solver.get_betas_r()

        # self.upper_beta = np.zeros((self.data.p, ))
        # self.upper_beta[self.allowed_support] = betas_on_allowed_support

        # self.upper_r = np.zeros((self.data.p, ))
        # self.upper_r[self.allowed_support] = r_on_allowed_support

        # self.upper_bound = min(upper_solver.loss_arr_child)

        (
            self.upper_beta,
            self.upper_r,
            self.upper_bound,
        ) = upper_solver.get_betas_r_loss()

        return self.upper_bound

    def lower_solve_brute_force(self, k, upper_solver):
        """Find the lower bound of the current node using brute force method

        Args:
            k (int): cardinality constraint
            upper_solver (CustomClass): solver that can find the best k-sparse solution via brute force

        Returns:
            float: loss of the best k-sparse solution
        """
        if self.data_isStored_on_allowed_support is False:
            self.store_data_on_allowed_support()

        allowed_supp_mask = self.zub == 1
        fixed_supp_mask = self.zlb == 1

        upper_solver.reset_fixed_supp_and_allowed_supp(
            fixed_supp_mask, allowed_supp_mask
        )
        upper_solver.get_sparse_sol_via_brute_force(k=k)

        (
            self.upper_beta,
            self.upper_r,
            self.upper_bound,
        ) = upper_solver.get_betas_r_loss()
        self.lower_solve_used_brute_force = True
        self.lower_bound = self.upper_bound

        return self.lower_bound

    def lower_solve_fast(self, k, upper_bound):
        """Solve the perspective relaxation of the current node through the Fast Solve method

        Args:
            k (int): cardinality constraint
            upper_bound (float): loss of the best k-sparse solution found so far

        Returns:
            float: a valid lower bound of the current node
        """
        if self.data_isStored_on_allowed_support is False:
            self.store_data_on_allowed_support()

        if len(self.fixed_support_on_allowed_support) >= k:
            self.lower_bound = upper_bound
            return self.lower_bound
        if self.inherit_parent_lower_sol is False:
            # self.lower_beta_on_allowed_support = np.linalg.solve(self.data.XTX_lambda2[self.allowed_support][:, self.allowed_support], self.data.XTy[self.allowed_support])
            # self.lower_beta_on_allowed_support = np.linalg.solve(self.XTX_lambda2_on_allowed_support, self.XTy_on_allowed_support)
            try:
                self.lower_beta_on_allowed_support = scipy.linalg.solve(
                    self.XTX_lambda2_on_allowed_support,
                    self.XTy_on_allowed_support,
                    assume_a="pos",
                )
            except:
                try:
                    self.lower_beta_on_allowed_support = np.linalg.solve(
                        self.XTX_lambda2_on_allowed_support, self.XTy_on_allowed_support
                    )
                except:
                    self.lower_beta_on_allowed_support = np.linalg.pinv(
                        self.XTX_lambda2_on_allowed_support
                    ).dot(self.XTy_on_allowed_support)

        self.lower_bound = self.calculate_lower_bound(
            self.lower_beta_on_allowed_support, k
        )

        return self.lower_bound

    def calculate_lower_bound(self, lower_beta_on_allowed_support, k):
        """Calculate the lower bound of the current node for the saddle point formulation

        Args:
            lower_beta_on_allowed_support (np.array): 1D array of coefficients for the saddle point formulation
            k (int): cardinality constraint

        Returns:
            float: a valid lower bound of the current node
        """

        lower_p_on_allowed_support = (
            2
            / self.data.lambda2_plus_smallest_eigenval
            * (
                self.XTy_on_allowed_support
                - self.XTX_minus_smallest_eigenval_on_allowed_support.dot(
                    lower_beta_on_allowed_support
                )
            )
        )

        lower_z_on_allowed_support = np.zeros((len(lower_beta_on_allowed_support),))

        lower_z_on_allowed_support[self.fixed_support_on_allowed_support] = 1
        lower_p_sq_on_allowed_support = lower_p_on_allowed_support ** 2
        largest_lower_p_sq_indices = np.argsort(
            -lower_p_sq_on_allowed_support[self.unfixed_support_on_allowed_support]
        )[: (k - len(self.fixed_support_on_allowed_support))]
        lower_z_on_allowed_support[
            self.unfixed_support_on_allowed_support[largest_lower_p_sq_indices]
        ] = 1

        term1 = self.XTX_minus_smallest_eigenval_on_allowed_support.dot(
            lower_beta_on_allowed_support
        ).dot(lower_beta_on_allowed_support)
        term2 = -2.0 * self.XTy_on_allowed_support.dot(lower_beta_on_allowed_support)
        term3 = self.data.lambda2_plus_smallest_eigenval * (
            lower_p_on_allowed_support.dot(lower_beta_on_allowed_support)
            - 1 / 4.0 * lower_z_on_allowed_support.dot(lower_p_sq_on_allowed_support)
        )

        lower_bound = term1 + term2 + term3

        return lower_bound

    def finetune_ADMM_rho(self, k, upper_bound, factor=10):
        """Finetune the step size of ADMM via two-way grid search

        Args:
            k (int): cardinality constraint
            upper_bound (float): loss of the best k-sparse solution found so far
            factor (int, optional): multiplicative factor when we perform grid search on the best step size. Defaults to 10.
        """
        # print("start finetuning ADMM rho")
        # print("at the beginning, rho is", self.data.rho_ADMM)
        original_rho = self.data.rho_ADMM

        _ = self.lower_solve(k, upper_bound)
        current_lower_bound = self.lower_solve_admm(k, upper_bound)

        total_factor = 1

        self.data.rho_ADMM = self.data.rho_ADMM / factor
        _ = self.lower_solve(k, upper_bound)
        tmp_lower_bound = self.lower_solve_admm(k, upper_bound)

        if tmp_lower_bound > current_lower_bound:
            factor = 1 / factor
            total_factor = total_factor * factor
            current_lower_bound = tmp_lower_bound

        while True:
            self.data.rho_ADMM = self.data.rho_ADMM * factor
            _ = self.lower_solve(k, upper_bound)
            tmp_lower_bound = self.lower_solve_admm(k, upper_bound)

            if tmp_lower_bound > current_lower_bound:
                total_factor = total_factor * factor
                current_lower_bound = tmp_lower_bound
            else:
                break
        self.data.rho_ADMM = original_rho * total_factor
        # print("at the end, rho is", self.data.rho_ADMM)

    def lower_solve_admm(self, k, upper_bound):
        """Solve the perspective relaxation of the current node through the ADMM method

        Args:
            k (int): cardinality constraint
            upper_bound (float): loss of the best k-sparse solution found so far

        Returns:
            float: a valid lower bound of the current node
        """
        if self.data_isStored_on_allowed_support is False:
            self.store_data_on_allowed_support()

        # print("upper_bound is", upper_bound)
        # print("initial lower bound is", self.lower_bound)

        lower_beta_on_allowed_support_k = self.lower_beta_on_allowed_support.copy()
        lower_p_on_allowed_support_k = (
            self.XTy_on_allowed_support
            - self.XTX_minus_smallest_eigenval_on_allowed_support.dot(
                lower_beta_on_allowed_support_k
            )
        )
        lower_q_on_allowed_support_k = np.zeros((len(lower_beta_on_allowed_support_k),))

        rho = self.data.rho_ADMM
        # print("rho inside ADMM is", rho)

        # step_1_Q = 2 / rho * np.eye(len(lower_beta_on_allowed_support_k)) + self.XTX_minus_smallest_eigenval_on_allowed_support
        step_1_Q = self.XTX_minus_smallest_eigenval_on_allowed_support.copy()
        np.fill_diagonal(step_1_Q, step_1_Q.diagonal() + 2 / rho)

        step_1_L, step_1_low = scipy.linalg.cho_factor(
            step_1_Q, overwrite_a=True, check_finite=False
        )
        del step_1_Q
        step_2_factor = 1 + 2 / (self.data.lambda2_plus_smallest_eigenval * rho)
        isotonic_clf = IsotonicRegression()

        k_minus_fixed = k - len(self.fixed_support_on_allowed_support)
        # print("k_minus_fixed is", k_minus_fixed)

        lower_beta_Q_on_allowed_support_k = np.zeros(
            (len(lower_beta_on_allowed_support_k),)
        )

        for _ in range(20):
            # step 1
            # lower_beta_on_allowed_support_k = np.linalg.solve(step_1_Q, self.XTy_on_allowed_support - lower_p_on_allowed_support_k - lower_q_on_allowed_support_k)
            lower_beta_on_allowed_support_k = scipy.linalg.cho_solve(
                (step_1_L, step_1_low),
                self.XTy_on_allowed_support
                - lower_p_on_allowed_support_k
                - lower_q_on_allowed_support_k,
                overwrite_b=True,
                check_finite=False,
            )

            lower_beta_Q_on_allowed_support_k = (
                2
                * self.XTX_minus_smallest_eigenval_on_allowed_support.dot(
                    lower_beta_on_allowed_support_k
                )
                + lower_p_on_allowed_support_k
                - self.XTy_on_allowed_support
            )

            # lower_beta_Q_on_allowed_support_k =  2 * lower_beta_on_allowed_support_k.dot(self.XTX_minus_smallest_eigenval_on_allowed_support)
            # lower_beta_Q_on_allowed_support_k += (lower_p_on_allowed_support_k - self.XTy_on_allowed_support)

            # np.dot(self.XTX_minus_smallest_eigenval_on_allowed_support_twice, lower_beta_on_allowed_support_k, out=lower_beta_Q_on_allowed_support_k)
            # lower_beta_Q_on_allowed_support_k += (lower_p_on_allowed_support_k - self.XTy_on_allowed_support)

            # step 2
            # lower_p_on_allowed_support_k = self.XTy_on_allowed_support - self.XTX_minus_smallest_eigenval_on_allowed_support.dot(lower_beta_on_allowed_support_k) - lower_q_on_allowed_support_k
            lower_p_on_allowed_support_k = (
                self.XTy_on_allowed_support
                - lower_beta_Q_on_allowed_support_k
                - lower_q_on_allowed_support_k
            )
            lower_p_on_allowed_support_k[
                self.fixed_support_on_allowed_support
            ] /= step_2_factor

            tmp_vec_x = np.abs(
                lower_p_on_allowed_support_k[self.unfixed_support_on_allowed_support]
            )
            tmp_vec_weight = np.ones((len(tmp_vec_x),))
            tmp_vec_y = tmp_vec_x.copy()
            top_k_minus_fixed_indicies = np.argpartition(tmp_vec_x, -k_minus_fixed)[
                -k_minus_fixed:
            ][:k_minus_fixed]
            tmp_vec_weight[top_k_minus_fixed_indicies] = step_2_factor
            tmp_vec_y[top_k_minus_fixed_indicies] /= step_2_factor

            lower_p_on_allowed_support_k[
                self.unfixed_support_on_allowed_support
            ] = np.sign(
                lower_p_on_allowed_support_k[self.unfixed_support_on_allowed_support]
            ) * isotonic_clf.fit_transform(
                X=tmp_vec_x, y=tmp_vec_y, sample_weight=tmp_vec_weight
            )

            # step 3
            # lower_q_on_allowed_support_k += (self.XTX_minus_smallest_eigenval_on_allowed_support.dot(lower_beta_on_allowed_support_k) + lower_p_on_allowed_support_k - self.XTy_on_allowed_support)
            lower_q_on_allowed_support_k += (
                lower_beta_Q_on_allowed_support_k
                + lower_p_on_allowed_support_k
                - self.XTy_on_allowed_support
            )

            # tmp_loss, _, _ = self.calculate_lower_bound(lower_beta_on_allowed_support_k, k)
            # print("loss becomes", tmp_loss)

            tmp_loss = self.calculate_lower_bound(lower_beta_on_allowed_support_k, k)
            # print("loss becomes", tmp_loss)

        tmp_lower_bound = self.calculate_lower_bound(lower_beta_on_allowed_support_k, k)
        # print("after ADMM, lower bound is {}".format(tmp_lower_bound))

        if tmp_lower_bound > self.lower_bound:
            self.lower_bound = tmp_lower_bound
            self.lower_beta_on_allowed_support = lower_beta_on_allowed_support_k.copy()
        # print("we are using {} as the final lower bound".format(self.lower_bound))
        # print()

        return self.lower_bound

def new_z(node, index):
    """Create two new z vectors by branching on the index-th variable

    Args:
        node (CustomClass): parent node
        index (int): index of the variable to branch on

    Returns:
        np.array: 1D array of left z vector
        np.array: 1D array of right z vector
    """
    new_zlb = node.zlb.copy()
    new_zub = node.zub.copy()
    new_zlb[index] = 1
    new_zub[index] = 0
    return new_zlb, new_zub


def branch(current_node, k):
    """Branch on the current node

    Args:
        current_node (CustomClass): parent node
        k (int): cardinality constraint

    Returns:
        CustomClass: left child node
        CustomClass: right child node
    """
    unfixed_and_allowed_support = current_node.allowed_support[
        current_node.unfixed_support_on_allowed_support
    ]
    nonzero_unfixed_and_allowed_support = unfixed_and_allowed_support[
        np.nonzero((current_node.upper_beta)[unfixed_and_allowed_support])
    ]

    delta_beta = -current_node.upper_beta[nonzero_unfixed_and_allowed_support]
    increase_in_loss = (
        current_node.data.half_Lipschitz[nonzero_unfixed_and_allowed_support]
        * delta_beta ** 2
        + 2 * current_node.upper_r[nonzero_unfixed_and_allowed_support] * delta_beta
        - 2 * current_node.data.XTy[nonzero_unfixed_and_allowed_support] * delta_beta
    )
    branching_variable = nonzero_unfixed_and_allowed_support[
        np.argmax(increase_in_loss)
    ]

    new_zlb, new_zub = new_z(current_node, branching_variable)
    # right_node = Node(current_node, new_zlb, current_node.zub.copy(), inherit_parent_upper_sol=(len(current_node.fixed_support_on_allowed_support) == k-1))
    right_node = Node(
        current_node,
        new_zlb,
        current_node.zub.copy(),
        inherit_parent_upper_sol=False,
        inherit_parent_lower_sol=True,
    )
    left_node = Node(current_node, current_node.zlb.copy(), new_zub)
    return left_node, right_node
